part of '../flutterx_live_data.dart';

mixin MediatorMixin<Y> on LiveData<Y> {
  final Map<LiveData, _Source> _sources = {};

  bool _reactive = false;

  /// If true, this LiveData will observe sources even if it's not currently active
  bool get reactive => _reactive;

  set reactive(bool value) {
    if (_reactive == value) return;
    _reactive = value;
    if (!isActive) value ? _plugAll() : _unplugAll();
  }

  @override
  void onActive() {
    super.onActive();
    if (!reactive) _plugAll();
  }

  @override
  void onInactive() {
    if (!reactive) _unplugAll();
    super.onInactive();
  }

  void addTransformSources(Iterable<LiveData> sources, FutureOr<Y> Function(List values) transform) =>
      _observeAllInternal(sources, (values) => postData(transform(values)), addSource);

  void addTransformSource<X>(LiveData<X> source, FutureOr<Y> Function(X) transform) =>
      addSource<X>(source, (value) => postData(transform(value)));

  void addLiveTransformSource<X>(LiveData<X> source, LiveData<Y> Function(X) transform) {
    LiveData<Y>? lastSource;
    addSource<X>(source, (value) {
      final newSource = transform(value);
      if (identical(lastSource, newSource)) return;
      if (lastSource != null) removeSource(lastSource!);
      addSource(lastSource = newSource, setData);
    });
  }

  void addSource<X>(LiveData<X> source, Observer<X> onChanged) {
    assert(!_sources.containsKey(source), 'source already registered');
    final wrapper = _Source<X>(source, onChanged);
    _sources[source] = wrapper;
    if (isActive || reactive) wrapper.plug();
  }

  void removeSource(LiveData source) => _sources.remove(source)?.unplug();

  void removeSources(Iterable<LiveData> sources) => sources.toList(growable: false).forEach(removeSource);

  void _plugAll() {
    for (final source in _sources.values) source.plug();
  }

  void _unplugAll() {
    for (final source in _sources.values) source.unplug();
  }
}

class _Source<X> {
  final LiveData<X> data;

  final Observer<X> onChanged;

  int lastVersion = _uninitialized;

  _Source(this.data, this.onChanged);

  void plug() => data.observeForever(_onChanged);

  void unplug() => data.removeObserver(_onChanged);

  void _onChanged(X value) {
    if (lastVersion < data._version) {
      lastVersion = data._version;
      onChanged(value);
    }
  }
}
